
// IDL-DODS interface.
// Written by Daniel J. Carr - Research Systems, Inc.
// Copyright 1999, UCAR
//
// IDL "LINKIMAGE" routines.
// 
// dods_connect()
// dods_var_p()
// dods_var_info()
// dods_next()
// get_dods_str()
// get_dods_byte()
// get_dods_int16()
// get_dods_uint16()
// get_dods_int32()
// get_dods_uint32()
// get_dods_float32()
// get_dods_float64()
// get_dods_array()
// get_dods_grid()
// dods_free()

// $Log: idl_dods.cc,v $
// Revision 1.4  1999/06/08 21:57:28  jimg
// Final fix for the String --> string changes. See line 355.
//
// Revision 1.3  1999/06/08 18:42:36  jimg
// Fixes for the String --> string fixes.
//
// Revision 1.2  1999/06/08 17:44:37  jimg
// Changed instances of String to string.
// Added suggestions for ways to shorten the code. These mostly come about
// because of the shift to string.
// Found several bugs in get_dods_str(...) which I tried to fix.
// Noted some omissions in the get_dods_array and get_dods_grid functions.
//
// Revision 1.1  1999/06/08 16:51:50  jimg
// Added

static char rcsid[]={"$Id: idl_dods.cc,v 1.4 1999/06/08 21:57:28 jimg Exp $"};

#include <stdio.h>

#include <Pix.h>
#include <string>

#include "BaseType.h"
#include "Connect.h"

// IDL include.
#include "export.h"

// dods_connect()
// Establish connection and return connect object and DDS object.
//    argv[0] - (Input ) url byte stream (null terminated).
//    argv[1] - (Input ) Constraint expression byte stream (null terminated).
//    argv[2] - (Output) Named IDL variable to recieve a pointer
//                       to the DDS object.
//    argv[3] - (Output) Named IDL variable to recieve the number
//                       of variables in the DDS.
//    argv[4] - (Input ) WWW_VERBOSE_ERRORS flag.
//    argv[5] - (Input ) ACCEPT_DEFLATE flag.
//    argv[6] - (Input ) GUI flag.
//    returns -          Pointer to Connect object.
//
// In version 3.0.x of DODS the progress dialog is not supported. In version
// 3.1 or 3.2 we will reintroduce this with a new, more robust,
// implementation. In this code I have disabled the GUI flag option (which
// activated the progress dialog feature). 6/8/99 jhrg
//
extern "C" IDL_VPTR dods_connect(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   bool www_verbose_errors = false;
   bool accept_deflate = false;
   bool gui_flag = false;
   DDS* dds;
   Gui* gui;

   if (argv[4]->value.ul == 0ul) www_verbose_errors = false;
   else www_verbose_errors = true;

   if (argv[5]->value.ul == 0ul) accept_deflate = false;
   else accept_deflate = true;

#if 0
   if (argv[6]->value.ul == 0ul) gui_flag = false;
   else gui_flag = true;
#endif

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   string url;
   url = string((char *) argv[0]->value.arr->data);
   string expr;
   expr = string((char *) argv[1]->value.arr->data);

   Connect* connect_obj = 0;
   connect_obj = new Connect(url, www_verbose_errors, accept_deflate);
   if (connect_obj == 0) return lReturn; // Connection failed.

#if 0
   gui = connect_obj->gui();
   gui->show_gui((int) gui_flag);
#endif

   dds = connect_obj->request_data(expr, gui_flag, false);
   if (dds == 0) return lReturn; // Request_data failed.

   lReturn->value.ul = (IDL_ULONG) connect_obj;
   argv[2]->value.ul = (IDL_ULONG) dds;
   argv[3]->value.ul = (IDL_ULONG) dds->num_var();

   return lReturn;
}


// dods_var_p()
// Get the pointer to a variable in the DDS.
//    argv[0] - (Input ) DDS pointer.
//    argv[1] - (Input ) Index of desired variable in the DDS.
//    returns -          Variable pointer.
//
extern "C" IDL_VPTR dods_var_p(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   DDS* dds;
   Pix q;
   int i;

   dds = (DDS *) argv[0]->value.ul;
   q = dds->first_var();

   for (i = 1; i < (int) argv[1]->value.ul; i++)
   {
      dds->next_var(q);
   }

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = (IDL_ULONG) dds->var(q);

   return lReturn;
}


// dods_var_info()
// Get information about a variable.
//    argv[0] - (Input ) Variable pointer.
//    argv[1] - (Output) Named IDL variable to receive DDS variable type code.
//                       Memory must be allocated for this IDL variable.
//    argv[2] - (Output) Named IDL variable to receive DDS variable's name.
//                       Memory must be allocated for this IDL variable.
//    argv[3] - (Output) Named IDL variable to receive the pseudo index of
//                       the first sub-variable, where applicable
//                       (for structure and sequence types).
//                       Memory must be allocated for this IDL variable.
//
extern "C" void dods_var_info(int argc, IDL_VPTR argv[])
{
   BaseType* var_p;
   Pix q;
   string var_name;
   unsigned int name_len;
   char* arr_ptr;
   int i;
   IDL_VPTR type;
   IDL_VPTR name_arr;
   IDL_VPTR name_arr2;
   IDL_VPTR pix;

   var_p = (BaseType *) argv[0]->value.ul;

   type = IDL_Gettmp(); // Allocate memory for output variable.
   type->type = IDL_TYP_ULONG;
   type->value.ul = var_p->type();
   IDL_VarCopy(type, argv[1]);

   var_name = var_p->name();
   name_len = var_name.length() + 1;
   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_BYTE, 1, (IDL_LONG *) &name_len,
                IDL_BARR_INI_ZERO, &name_arr);
   for (i = 0; i < name_len-1; i++)
   {
      name_arr->value.arr->data[i] = (UCHAR) var_name[i];
   }
#if 0
   // Suggestion: rather than use a loop and length of the name, use string's
   // c_str() and strcpy(). This following line does all that the code from
   // var_name = ... through the for loop does. Of course, when Dan wrote
   // this we were still using GNU String. 6/8/99 jhrg
   strcpy(name_arr->value.arr->data, var_p->name().c_str());
#endif
   name_arr2 = IDL_CvtString(1, &name_arr, NULL);
   IDL_VarCopy(name_arr2, argv[2]);
   if (name_arr2 != name_arr) IDL_DELTMP(name_arr);

   pix = IDL_Gettmp(); // Allocate memory for output variable.
   pix->type = IDL_TYP_ULONG;
   pix->value.ul = 0ul;
   switch (var_p->type())
   {
      case dods_structure_c:
         pix->value.ul = (IDL_ULONG) ((Structure *) var_p)->first_var();
      break;

      case dods_sequence_c:
         pix->value.ul = (IDL_ULONG) ((Sequence *) var_p)->first_var();
      break;

      default:
      break;
   }
   IDL_VarCopy(pix, argv[3]);
}


// dods_seq()
// Read the next instance of a sequence (deserialize).
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    returns -          Status (1 if more instances left to read,
//                       0 if done).
//
extern "C" IDL_VPTR dods_seq(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   Connect* connect_obj;
   DDS* dds;
   BaseType* var_p;
   bool status;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   status = var_p->deserialize(connect_obj->source(), dds);

   if (status == true) lReturn->value.ul = (IDL_ULONG) 1;
   return lReturn;
}


// dods_next()
// Get the pseudo index of the next variable in a structure or sequence.
//    argv[0] - (Input ) Variable pointer.
//    argv[1] - (I/O   ) Pseudo index.   Memory already allocated.
//    argv[2] - (Output) Named IDL variable to receive the pointer to the
//                       DDS variable at the current pseudo index.
//                       Memory already allocated.
//
extern "C" void dods_next(int argc, IDL_VPTR argv[])
{
   BaseType* var_p;
   Pix q;
   BaseType* sub_var_p;

   var_p = (BaseType *) argv[0]->value.ul;
   q = (Pix) argv[1]->value.ul;

   switch (var_p->type())
   {
      case dods_structure_c:
         sub_var_p = ((Structure *) var_p)->var(q);
         ((Structure *) var_p)->next_var(q);
      break;

      case dods_sequence_c:
         sub_var_p = ((Sequence *) var_p)->var(q);
         ((Sequence *) var_p)->next_var(q);
      break;

      default:
         sub_var_p = 0;
         q = 0;
      break;
   }

   argv[1]->value.ul = (IDL_ULONG) q;
   argv[2]->value.ul = (IDL_ULONG) sub_var_p;
}


// get_dods_str()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_str(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim;
   int i;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;
   // Fix: See further down where the call to malloc, buf2val and free are
   // changed. This was necessary due to changes in the DODS DAP software.
   // 6/8/99 jhrg
#if 0
   UCHAR* str_ptr;
#endif
   string *str_ptr;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   n_bytes = var_p->width();
   if (n_bytes <= 0) return lReturn;

   // Allocate memory for output variable.
   dim = n_bytes + 1;

   // There are some serious problems here. malloc is used but inside
   // Str::buf2val(...) those bytes are going to be deleted using delete().
   // That's going to cause problems. If `str_ptr' is null, then buf2val(...)
   // will allocate storage for you. 6/8/99 jhrg
   //
   // Suggestion: Change the malloc for str_ptr to str_ptr = 0; I'm going to
   // do this one. 6/8/99 jhrg
#if 0
   str_ptr = (UCHAR *) malloc(dim);
#endif
   string str;
   str_ptr = &str;
   n_bytes = var_p->buf2val((void **) &str_ptr);

   arr_ptr = IDL_MakeTempArray(IDL_TYP_BYTE, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);

   IDL_VarCopy(arr_data, argv[3]);

   // Suggestion: Assuming data has enough space use strcpy here. 6/8/99 jhrg
   for (i = 0; i < n_bytes; i++)
   {
      argv[3]->value.arr->data[i] = str[i];
   }
   // Fix: This must be delete. 6/8/99 jhrg
#if 0
   free(str_ptr);
#endif
   delete(str_ptr);

   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_byte()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_byte(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_BYTE, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_int16()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_int16(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_INT, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_uint16()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_uint16(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_UINT, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_int32()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_int32(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_LONG, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_uint32()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_uint32(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_ULONG, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_float32()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_float32(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_FLOAT, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_float64()
// Get the data from a DDS variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_float64(int argc, IDL_VPTR argv[])
{
   IDL_VPTR lReturn;
   BaseType* var_p;
   char* arr_ptr;
   unsigned int n_bytes;
   int dim=1;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = 0ul;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (BaseType *) argv[2]->value.ul;

   // Allocate memory for output variable.
   arr_ptr = IDL_MakeTempArray(IDL_TYP_DOUBLE, (int) 1, 
                (IDL_LONG *) &dim, IDL_BARR_INI_ZERO, &arr_data);
   IDL_VarCopy(arr_data, argv[3]);

   n_bytes = var_p->buf2val((void **) &argv[3]->value.arr->data);
   if (n_bytes <= 0) return lReturn;

   lReturn->value.ul = 1;
   return lReturn;
}


// get_dods_array()
// Get the data from a DDS array variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the data.
//                       Memory must be allocated.
//    argv[4] - (Output) Named IDL variable (array) to receive dimension names.
//                       Memory already allocated as a (65x8) byte array.
//    argv[5] - (Output) Named IDL variable (array) to receive the dimension
//                       start/stride/stop values.   Memory already allocated
//                       as a (3x8) long int array.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_array(int argc, IDL_VPTR argv[])
{
   unsigned int ret_val=1;
   IDL_VPTR lReturn;
   Array* var_p;
   BaseType* dim_var;
   BaseType* cell_var;
   unsigned int n_dim;
   int dims[8];
   string dim_name;
   unsigned int name_len;
   Pix q;
   char* arr_ptr;
   unsigned int n_bytes;
   int i, j;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   UCHAR* uchar_p;
   IDL_LONG* long_p;
   IDL_VPTR arr_data;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = ret_val;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (Array *) argv[2]->value.ul;

   n_dim = var_p->dimensions();
   if (n_dim <= 8)
   {
      q = var_p->first_dim();

      if (q != 0ul)
      {
         for (i = 0; i < n_dim; i++)
         {
	     // Suggestion: use strncpy(...) here. 6/8/99 jhrg
            dim_name = var_p->dimension_name(q);
            name_len = dim_name.length();
            if (name_len >= 64) name_len = 64;
            for (j = 0; j < name_len; j++)
            {
               uchar_p = (UCHAR *) &argv[4]->value.arr->data[0] + ((65*i) + j);
               *uchar_p = (UCHAR) dim_name[j];
            }

            long_p = (IDL_LONG *) &argv[5]->value.arr->data[0] + (3*i);
            *long_p = (IDL_LONG) var_p->dimension_start(q);
            long_p++;
            *long_p = (IDL_LONG) var_p->dimension_stride(q);
            long_p++;
            *long_p = (IDL_LONG) var_p->dimension_stop(q);

            dims[i] = var_p->dimension_size(q);

            var_p->next_dim(q);
            if (q == 0ul) i = n_dim + 1;
         }
      }

      cell_var = var_p->var(0);

      // Allocate memory and copy data to output variable.

      switch (cell_var->type())
      {
         case dods_byte_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_BYTE, (int) n_dim, 
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         case dods_int16_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_INT, (int) n_dim,
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         case dods_uint16_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_UINT, (int) n_dim,
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         case dods_int32_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_LONG, (int) n_dim,
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         case dods_uint32_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_ULONG, (int) n_dim,
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         case dods_float32_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_FLOAT, (int) n_dim,
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         case dods_float64_c:
            arr_ptr = IDL_MakeTempArray(IDL_TYP_DOUBLE, (int) n_dim,
                         (IDL_LONG *) dims, IDL_BARR_INI_NOP, &arr_data);
            n_bytes = var_p->buf2val((void **) &arr_data->value.arr->data);
            IDL_VarCopy(arr_data, argv[3]);
         break;
         default:
	   // Question: What about an array of strings? 6/8/99 jhrg
         break;
      }
   }
   else ret_val = 0;

   lReturn->value.ul = ret_val;

   return lReturn;
}


// get_dods_grid()
// Get the data from a DDS grid variable.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//    argv[2] - (Input ) Variable pointer.
//    argv[3] - (Output) Named IDL variable to receive the grid variable pointer.
//                       IDL variable must be already defined as an unsigned long.
//    argv[4] - (Output) Named IDL variable (array) to receive dimension names.
//                       Memory already allocated as a (64x8) byte array.
//    argv[5] - (Output) Named IDL variable to receive the 1st dimension array.
//                       Memory must be allocated.
//    argv[6-12] - 
//              (Output) Named IDL variables to receive the 2nd through 8th
//                       dimension arrays (if applicable).   Memory must be allocated.
//    returns -          Status (1=ok, 0=fail).
//
extern "C" IDL_VPTR get_dods_grid(int argc, IDL_VPTR argv[])
{
   unsigned int ret_val=1;
   IDL_VPTR lReturn;
   Grid* var_p;
   Array* grid_var;
   Vector* dim_var;
   unsigned int n_dim;
   int dims[8];
   int dim;
   string dim_name;
   unsigned int name_len;
   Pix q;
   char* arr_ptr;
   unsigned int n_bytes;
   int i, j;
   Connect* connect_obj;
   DDS* dds;
   bool status;
   IDL_VPTR arr_data;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;
   var_p = (Grid *) argv[2]->value.ul;

   grid_var = (Array *) var_p->array_var();
   n_dim = grid_var->dimensions();
   if (n_dim <= 8)
   {
      q = var_p->first_map_var();

      if (q != 0ul)
      {
         for (i = 0; i < n_dim; i++)
         {
	     // Suggestion: replace with strncpy(...). 6/8/99 jhrg
            dim_var = (Vector *) var_p->map_var(q);
            dim_name = dim_var->name();
            name_len = dim_name.length();
            if (name_len >= 64) name_len = 64;
            for (j = 0; j < name_len-1; j++)
            {
               argv[4]->value.arr->data[j,i] = (UCHAR) dim_name[j];
            }

            // Allocate memory and copy data to output dimension variables.

            dim = dim_var->length();
            switch (dim_var->type())
            {
               case dods_byte_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_BYTE, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               case dods_int16_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_INT, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               case dods_uint16_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_UINT, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               case dods_int32_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_LONG, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               case dods_uint32_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_ULONG, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               case dods_float32_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_FLOAT, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               case dods_float64_c:
                  arr_ptr = IDL_MakeTempArray(IDL_TYP_DOUBLE, (int) 1, 
                               (IDL_LONG *) &dim, IDL_BARR_INI_NOP, &arr_data);
                  n_bytes = dim_var->buf2val((void **) &arr_data->value.arr->data);
                  IDL_VarCopy(arr_data, argv[i+5]);
               break;
               default:
		 // Question: What about otehr types (strings, etc.)? 6/8/99
		 // jhrg 
               break;
            }
            var_p->next_map_var(q);
            if (q == 0ul) i = n_dim + 1;
         }
      }

      // Return grid data pointer.
      argv[3]->value.ul = (IDL_ULONG) grid_var;
   }
   else ret_val = 0;

   lReturn = IDL_Gettmp(); // Allocate memory for return variable.
   lReturn->type = IDL_TYP_ULONG;
   lReturn->value.ul = ret_val;

   return lReturn;
}


// dods_free()
// Delete and free the Connect and DDS objects.
//    argv[0] - (Input ) Pointer to the connect object.
//    argv[1] - (Input ) Pointer to the DDS object.
//
extern "C" IDL_VPTR dods_free(int argc, IDL_VPTR argv[])
{
   Connect* connect_obj;
   DDS* dds;

   connect_obj = (Connect *) argv[0]->value.ul;
   dds = (DDS *) argv[1]->value.ul;

   delete dds;
   // I think that removing the gui stuff where Connect was instantiated
   // should fix the nasty problems with core dumps/seg faults when
   // connect_obj is deleted. 6/8/99 jhrg
   //delete connect_obj;
}
